package be.iminds.ilabt.jfed.restapi;

import be.iminds.ilabt.jfed.fedmon.util.Clock;
import be.iminds.ilabt.jfed.restapi.TestDB;
import be.iminds.ilabt.jfed.restapi.client.JFedRestApiClient;
import be.iminds.ilabt.jfed.restapi.client.JFedRestApiClientConfig;
import be.iminds.ilabt.jfed.restapi.client.JFedRestApiClientImpl;
import be.iminds.ilabt.jfed.restapi.client.JFedRestApiClientPropertiesConfig;
import be.iminds.ilabt.jfed.restapi.json.Experiment;
import be.iminds.ilabt.jfed.restapi.json.ExperimentBuilder;
import be.iminds.ilabt.jfed.restapi.json.ExperimentTestCollection;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import be.iminds.ilabt.jfed.util.KeyUtil;
import be.iminds.ilabt.util.jsonld.test.CommonTest;
import io.dropwizard.testing.junit.DropwizardAppRule;
import org.junit.*;

import java.io.IOException;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.*;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;

import static org.mockito.Mockito.mock;

/**
 * ExperimentResourceIT: temporary test using SSL server and credential.
 */
public class ExperimentResourceIT {
    @ClassRule
    public static final DropwizardAppRule<JFedRestApiConfiguration> RULE =
            new DropwizardAppRule<JFedRestApiConfiguration>(JFedRestApiApplication.class, "../jfed-rest-api-TESTING-SSL.yml");

//    protected static ExperimentTestCollection experimentTestCollection;

    protected JFedRestApiClient client;

    private static TestDB testDB;
    private ExperimentTestCollection experimentTestCollection;

    @BeforeClass
    public static void setUp() throws Exception {
        setUp(RULE);
    }
    public static void setUp(DropwizardAppRule<JFedRestApiConfiguration> RULE) throws Exception {
        Clock.set(new Clock.DefaultTestTimeProvider());

        assert RULE != null;

        testDB = new TestDB(RULE);
        testDB.recreateTableStructure();
        testDB.fillAllTables();

        assert RULE.getConfiguration().isTest();
    }

    @AfterClass
    public static void tearDown() throws Exception {
        Clock.reset();
    }

    @Before
    public void beforeMethod() throws Exception {
        assert RULE.getConfiguration().isTest();

        JFedTrustStore jFedTrustStore = new JFedTrustStore();
        String serverPem = IOUtils.fileToString("/home/wim/git/jfed-rest-api/generate-test-certs/testing-jfed-rest-api/ca/intermediate/certs/ca-chain.cert.pem");
//        List<X509Certificate> serverCerts = KeyUtil.pemToX509CertificateChain(serverPem);
        jFedTrustStore.addTrustedPemCertificate(serverPem);

        String userLoginPem = IOUtils.fileToString("/home/wim/git/jfed-rest-api/generate-test-certs/testing-jfed-rest-api/ca/client1.pem");
        X509Certificate userCert = KeyUtil.pemToX509Certificate(userLoginPem);
        PrivateKey userPrivKey = KeyUtil.pemToAnyPrivateKey(userLoginPem, null);

        Properties properties = new Properties();
        properties.setProperty("jfed_restapi_client_url_read_base", String.format("https://localhost:%d/api/", RULE.getLocalPort()));
        properties.setProperty("jfed_restapi_client_url_write_base", String.format("https://localhost:%d/api/", RULE.getLocalPort()));

        JFedRestApiClientConfig clientConfig =
                new JFedRestApiClientPropertiesConfig(
                        properties,
                        jFedTrustStore,
                        Collections.singletonList(userCert),
                        userPrivKey);
        client = new JFedRestApiClientImpl(clientConfig);
        assert client != null;

        experimentTestCollection = new ExperimentTestCollection();
        testDB.refillDB();
    }

    @Test(expected = JFedRestApiClient.JFedRestApiClientException.class)
    public void testGetNonExisting() throws Exception {
        long maxId = experimentTestCollection.getAllBuilders().stream()
                .map(ExperimentBuilder::getId)
                .filter(Objects::nonNull)
                .max(Comparator.naturalOrder())
                .orElse(0l);
        long nonExistingId = maxId+1;
        System.out.println("maxId="+maxId+" nonExistingId="+nonExistingId);

        Experiment actual = client.getById(Experiment.class, nonExistingId).get();
        //will not get here: error is thrown
        assertThat(actual, is(nullValue()));
    }

    @Test
    public void testGetExistingNotFull() throws Exception {
        Experiment expected = experimentTestCollection.getByIndex(0);
        assert expected.getId() != null;
        System.out.println("looking up id="+expected.getId());

        Experiment actual = client.getById(Experiment.class, expected.getId()).get();

        assertThat(actual, is(notNullValue()));

        System.out.println("Got result Experiment: "+actual);

        expected = new ExperimentBuilder(expected).setRequestRspec(null).setSpeaksForCredential(null).create();

        experimentTestCollection.assertSame(actual, expected,
                CommonTest.ComparePrecision.FULL,
                CommonTest.ComparePrecision.FULL);
    }

    @Test
    public void testGetExistingFull() throws Exception {
        Experiment expected = experimentTestCollection.getByIndex(0);
        assert expected.getId() != null;
        System.out.println("looking up id="+expected.getId());

        Experiment actual = client.getById(Experiment.class, expected.getId(), true/*full*/).get();

        assertThat(actual, is(notNullValue()));

        System.out.println("Got result Experiment: "+actual);

        experimentTestCollection.assertSame(actual, expected,
                CommonTest.ComparePrecision.FULL,
                CommonTest.ComparePrecision.FULL);
    }

    public Experiment createExperimentForTest() throws IOException {
        String credentialsXmlString = IOUtils.fileToString("/home/wim/git/jfed-rest-api/credential-ftester_speaksfor_wvdemeer.xml");

        ExperimentBuilder experimentBuilder = new ExperimentBuilder();
//        experimentBuilder.setName("test1");
        experimentBuilder.setProject("fed4fire");
        experimentBuilder.setRequestRspec("<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" " +
                "      type=\"request\" " +
                "      generated_by=\"jFed RSpec Editor\" generated=\"2016-05-24T11:35:29.953+02:00\" " +
                "      xmlns:emulab=\"http://www.protogeni.net/resources/rspec/ext/emulab/1\" " +
                "      xmlns:jfedBonfire=\"http://jfed.iminds.be/rspec/ext/jfed-bonfire/1\" " +
                "      xmlns:delay=\"http://www.protogeni.net/resources/rspec/ext/delay/1\" " +
                "      xmlns:jfed-command=\"http://jfed.iminds.be/rspec/ext/jfed-command/1\" " +
                "      xmlns:client=\"http://www.protogeni.net/resources/rspec/ext/client/1\" " +
                "      xmlns:jfed-ssh-keys=\"http://jfed.iminds.be/rspec/ext/jfed-ssh-keys/1\" " +
                "      xmlns:jfed=\"http://jfed.iminds.be/rspec/ext/jfed/1\" " +
                "      xmlns:sharedvlan=\"http://www.protogeni.net/resources/rspec/ext/shared-vlan/1\" " +
                "      xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " +
                "      xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \">\n" +
                "  <node client_id=\"node0\" exclusive=\"true\" component_manager_id=\"urn:publicid:IDN+wall2.ilabt.iminds.be+authority+cm\">\n" +
                "    <sliver_type name=\"please-fail\"/>\n" +
                "  </node>\n" +
                "</rspec>");
        experimentBuilder.setSpeaksForCredential(credentialsXmlString);

        return experimentBuilder.create();
    }

    @Test
    public void testPrintExperimentRequest() throws Exception {
        Experiment createRequest = createExperimentForTest();

        System.out.println("Experiment request: " + createRequest);
    }

    //TODO implement
    //too much to mock for testCreate(). Will not do it initially
//    @Test
//    public void testCreate() throws Exception {
//        Experiment createRequest = createExperimentForTest();
//
//        Experiment created = client.create(createRequest);
//
//        System.out.println("Got created Experiment: "+created);
//
//
//        int count = 0;
//        boolean done = false;
//        while (!done) {
//            Thread.sleep(1000l);
//
//            Optional<Experiment> statusUpdate = client.getById(Experiment.class, created.getId());
//
//            assertThat(statusUpdate.isPresent(), is(true));
//
//            System.out.println("Got statusUpdate Experiment: " + statusUpdate.get());
//
//            switch (statusUpdate.get().getStatus()) {
//                case STARTING: { done = false; break; }
//                case FAILED: { done = true; break; }
//                case NONEXISTING: { done = true; break; }
//                case READY: { done = true; break; }
//                default: throw new RuntimeException("Unsupported status: "+statusUpdate.get().getStatus());
//            }
//
//            if (done) {
//                assertThat(statusUpdate.get().getStatus(), is(Experiment.Status.FAILED));
//            }
//
//            count++;
//            assertThat(count, is(lessThan(20)));
//        }
//    }
}
